@@ -41,6 +41,7 @@
 #![allow(clippy::result_large_err)] // temporary workaround for arti#587
 #![allow(clippy::needless_raw_string_hashes)] // complained-about code is fine, often best
 #![allow(clippy::needless_lifetimes)] // See arti#1765
+#![allow(mismatched_lifetime_syntaxes)] // temporary workaround for arti#2060
 //! <!-- @@ end lint list maintained by maint/add_warning @@ -->
 
 use std::path::Path;
@@ -120,10 +121,7 @@ impl LockFileGuard {
         if os::lockfile_has_path(&self.locked, path)? {
             std::fs::remove_file(path)
         } else {
-            Err(std::io::Error::new(
-                std::io::ErrorKind::Other,
-                MismatchedPathError {},
-            ))
+            Err(std::io::Error::other(MismatchedPathError {}))
         }
     }
 }
